#include "ViewerFoldingList.h"

FXDEFMAP(ViewerFoldingList) ViewerFoldingListMap[] = {
  FXMAPFUNC(SEL_LEFTBUTTONPRESS, 0, ViewerFoldingList::onLeftBtnPress), 
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS, 0, ViewerFoldingList::onMiddleBtnPress), 
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE, 0, ViewerFoldingList::onLeftBtnRelease), 
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE, 0, ViewerFoldingList::onMiddleBtnRelease), 
  FXMAPFUNC(SEL_MOTION, 0, ViewerFoldingList::onMotion), 
  FXMAPFUNC(SEL_BEGINDRAG, 0, ViewerFoldingList::onBeginDrag), 
  FXMAPFUNC(SEL_ENDDRAG, 0, ViewerFoldingList::onEndDrag), 
  FXMAPFUNC(SEL_DRAGGED, 0, ViewerFoldingList::onDragged), 
  FXMAPFUNC(SEL_DND_MOTION, 0, ViewerFoldingList::onDNDMotion), 
  FXMAPFUNC(SEL_DND_DROP, 0, ViewerFoldingList::onDNDDrop), 
  FXMAPFUNC(SEL_DND_REQUEST, 0, ViewerFoldingList::onDNDRequest)
};

FXIMPLEMENT(ViewerFoldingList, FXFoldingList, ViewerFoldingListMap, ARRAYNUMBER(ViewerFoldingListMap))

// Drag and drop type for the list items
const FXchar ViewerFoldingList::foldingItemTypeText[] = "VIEWERFOLDINGITEM";
FXDragType ViewerFoldingList::foldingItemType = 0;

ViewerFoldingList::ViewerFoldingList(FXComposite *p, FXObject* tgt, FXSelector sel, FXuint opts, FXint x, FXint y, FXint w, FXint h)
: FXFoldingList(p, tgt, sel, opts, x, y, w, h), 
  dragmode_(DND_NONE)
{
}

void ViewerFoldingList::create()
{
  FXFoldingList::create();

  if(!foldingItemType)
  {
    foldingItemType = getApp()->registerDragType(foldingItemTypeText);
  }

  dropEnable();
}

long ViewerFoldingList::onLeftBtnPress(FXObject* sender, FXSelector sel, void* ptr)
{
  FXFoldingList::onLeftBtnPress(sender, sel, ptr);

  dragmode_ = DND_TRYDRAG;

  grab();

  return 1;
}

long ViewerFoldingList::onMiddleBtnPress(FXObject* sender, FXSelector sel, void* ptr)
{
  FXFoldingList::onMiddleBtnPress(sender, sel, ptr);

  dragmode_ = DND_TRYDRAG;

  grab();

  return 1;
}

long ViewerFoldingList::onLeftBtnRelease(FXObject* sender, FXSelector sel, void* ptr)
{
  FXFoldingList::onLeftBtnRelease(sender, sel, ptr);

  if(dragmode_ == DND_DRAG)
  {
    handle(this, MKUINT(0, SEL_ENDDRAG), ptr);
  }

  dragmode_ = DND_NONE;

  ungrab();

  return 0;
}

long ViewerFoldingList::onMiddleBtnRelease(FXObject* sender, FXSelector sel, void* ptr)
{
  FXFoldingList::onLeftBtnRelease(sender, sel, ptr);

  if(dragmode_ == DND_DRAG)
  {
    handle(this, MKUINT(0, SEL_ENDDRAG), ptr);
  }

  dragmode_ = DND_NONE;

  ungrab();

  return 1;
}

long ViewerFoldingList::onMotion(FXObject* sender, FXSelector sel, void* ptr)
{
  FXEvent* event = (FXEvent*)ptr;

  switch(dragmode_)
  {
    case DND_DRAG:
      handle(this, MKUINT(0, SEL_DRAGGED), ptr);
      break;
    case DND_TRYDRAG:
      if(event->moved)
      {
        dragmode_ = DND_NONE;

        if(handle(this, MKUINT(0, SEL_BEGINDRAG), ptr))
        {
          dragmode_ = DND_DRAG;
        }
      }
      break;
  }

  // Drag mode may have been changed by one of the previous cases
  if(dragmode_ == DND_NONE)
  {
    FXFoldingList::onMotion(sender, sel, ptr);
  }

  return 1;
}

long ViewerFoldingList::onBeginDrag(FXObject*, FXSelector, void*)
{
  beginDrag(&foldingItemType, 1);

  setDragCursor(getApp()->getDefaultCursor(DEF_DNDMOVE_CURSOR));

  return 1;
}

long ViewerFoldingList::onEndDrag(FXObject*, FXSelector, void*)
{
  endDrag((didAccept() != DRAG_REJECT));

  setDragCursor(getApp()->getDefaultCursor(DEF_ARROW_CURSOR));

  return 1;
}

long ViewerFoldingList::onDragged(FXObject*, FXSelector, void* ptr)
{
  FXEvent* event = (FXEvent*)ptr;
  FXDragAction action = DRAG_MOVE;

  handleDrag(event->root_x, event->root_y, action);

  if(didAccept() != DRAG_REJECT)
  {
    setDragCursor(getApp()->getDefaultCursor(DEF_DNDMOVE_CURSOR));
  }
  else
  {
    setDragCursor(getApp()->getDefaultCursor(DEF_DNDSTOP_CURSOR));
  }

  return 1;
}

long ViewerFoldingList::onDNDMotion(FXObject*, FXSelector, void*)
{
  FXDragAction action;

  // Correct drop type
  if(offeredDNDType(FROM_DRAGNDROP, foldingItemType))
  {
    action = inquireDNDAction();

    // Check for legal DND action
    if(action == DRAG_MOVE)
    {
      acceptDrop(DRAG_ACCEPT);
    }

    return 1;
  }

  // Didn't handle it here
  return 0;
}

long ViewerFoldingList::onDNDDrop(FXObject*, FXSelector, void* ptr)
{
  if(target && target->handle(this, MKUINT(message, SEL_DND_DROP), ptr)) return 1;
  return 0;
}

long ViewerFoldingList::onDNDRequest(FXObject*, FXSelector, void* ptr)
{
  if(target && target->handle(this, MKUINT(message, SEL_DND_REQUEST), ptr)) return 1;
  return 0;
}
