#include <typeinfo>
#include <osg/Switch>
#include <osg/PositionAttitudeTransform>
#include <osgDB/ReadFile>
#include <osgDB/WriteFile>
#include <osgUtil/Optimizer>
#include <osgUtil/SceneView>
#include <osgGA/TrackballManipulator>
#include <osgGA/StateSetManipulator>
#include "RatioViewport.h"
#include "ViewportDialog.h"
#include "ViewerFoldingList.h"
#include "ViewerWindow.h"

#include <osg/ShapeDrawable>

#define TITLE "Scene Graph Editor"

FXDEFMAP(ViewerWindow) ViewerWindowMap[] = {
  FXMAPFUNC(SEL_CLOSE, 0, ViewerWindow::onCmdQuit),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_QUIT, ViewerWindow::onCmdQuit),
  FXMAPFUNC(SEL_UPDATE, ViewerWindow::ID_TITLEBAR, ViewerWindow::onUpdTitleBar),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_LOADSCENE, ViewerWindow::onCmdLoadScene),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_SAVESCENE, ViewerWindow::onCmdSaveScene),
  FXMAPFUNC(SEL_UPDATE, ViewerWindow::ID_SAVESCENE, ViewerWindow::onUpdSaveScene),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_SAVESCENEAS, ViewerWindow::onCmdSaveSceneAs),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_MANAGEVIEWPORTS, ViewerWindow::onCmdManageViewports),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_DRAWVIEWPORT, ViewerWindow::onCmdDrawViewport),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_RESETCAMERAS, ViewerWindow::onCmdResetCameras),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_SETBACKCOLOR, ViewerWindow::onCmdSetBackColor),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_SETAMBIENTCOLOR, ViewerWindow::onCmdSetAmbientColor),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS, ViewerWindow::ID_CANVAS, ViewerWindow::onLeftBtnPress), 
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE, ViewerWindow::ID_CANVAS, ViewerWindow::onLeftBtnRelease), 
  FXMAPFUNC(SEL_MOTION, ViewerWindow::ID_CANVAS, ViewerWindow::onMotion), 
  FXMAPFUNC(SEL_SELECTED, ViewerWindow::ID_LIST, ViewerWindow::onCmdList),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_NODENAME, ViewerWindow::onCmdNodeName),
  FXMAPFUNC(SEL_COMMAND, ViewerWindow::ID_SWITCHNODE, ViewerWindow::onCmdSwitchNode),
  FXMAPFUNCS(SEL_COMMAND, ViewerWindow::ID_TRANSLATION_X, ViewerWindow::ID_SCALE_Z, ViewerWindow::onCmdPositionAttitudeTransform),
  FXMAPFUNC(SEL_REPLACED, ViewerWindow::ID_GEOMTABLE, ViewerWindow::onCmdGeometryTable),
  FXMAPFUNC(SEL_CHORE, ViewerWindow::ID_GEOMCHORE, ViewerWindow::onGeometryChore),
  FXMAPFUNC(SEL_CHORE, ViewerWindow::ID_DRAWCHORE, ViewerWindow::onDrawChore),
  FXMAPFUNC(SEL_DND_DROP, ViewerWindow::ID_LIST, ViewerWindow::onDNDDrop),
  FXMAPFUNC(SEL_DND_REQUEST, ViewerWindow::ID_LIST, ViewerWindow::onDNDRequest),
};

FXIMPLEMENT(ViewerWindow, FXMainWindow, ViewerWindowMap, ARRAYNUMBER(ViewerWindowMap))

ViewerWindow::ViewerWindow(FXApp* app)
: FXMainWindow(app, TITLE, NULL, NULL, DECOR_ALL, 0, 0, 850, 600),
  dirty_(false),
  drawingViewport_(false),
  loadScenePath_(FXSystem::getCurrentDirectory()),
  currentFile_(FXString::null),
  root_(new osg::Group()),
  currentItem_(NULL)
{
  setTarget(this);
  setSelector(ID_TITLEBAR);

  root_->setName("Root");

  // Sites for docking
  FXDockSite* topdock = new FXDockSite(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X);
  new FXDockSite(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X);
  new FXDockSite(this,LAYOUT_SIDE_LEFT|LAYOUT_FILL_Y);
  new FXDockSite(this,LAYOUT_SIDE_RIGHT|LAYOUT_FILL_Y);

  // Menu bar
  FXToolBarShell* dragshell1=new FXToolBarShell(this,FRAME_RAISED);
  FXMenuBar* menubar=new FXMenuBar(topdock,dragshell1,LAYOUT_DOCK_SAME|LAYOUT_SIDE_TOP|LAYOUT_FILL_X|FRAME_GROOVE);
  new FXToolBarGrip(menubar,menubar,FXMenuBar::ID_TOOLBARGRIP,TOOLBARGRIP_DOUBLE);

  // Status bar
  statusbar_=new FXStatusBar(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER);


  // File and Toolbar items
  // File menu
  filemenu_=new FXMenuPane(this);
  new FXMenuCommand(filemenu_, "&Open...\tCtrl-O\tLoad 3D scene data from a file", NULL, this, ID_LOADSCENE);
  new FXMenuCommand(filemenu_, "&Save\tCtrl-S\tSave 3D scene data to file", NULL, this, ID_SAVESCENE);
  new FXMenuCommand(filemenu_, "Save &As...\tCtrl-Shift-S\tSave 3D scene data to a file", NULL, this, ID_SAVESCENEAS);
  new FXMenuSeparator(filemenu_);
  new FXMenuCommand(filemenu_, "&Quit\tCtrl-Q\tExit application", NULL, this, ID_QUIT);
  new FXMenuTitle(menubar, "&File", NULL, filemenu_);

  // View menu
  viewmenu_=new FXMenuPane(this);
  new FXMenuCommand(viewmenu_, "&Manage Viewports...\tCtrl-V\tDisplay the viewport management dialog", NULL, this, ID_MANAGEVIEWPORTS);
  new FXMenuCommand(viewmenu_, "Create Viewport\tV\tCreate a viewport by dragging the mouse on the canvas", NULL, this, ID_DRAWVIEWPORT);
  new FXMenuSeparator(viewmenu_);
  new FXMenuCommand(viewmenu_, "&Reset Cameras\tR\tReset the camera(s) to the default position", NULL, this, ID_RESETCAMERAS);
  new FXMenuSeparator(viewmenu_);
  new FXMenuCommand(viewmenu_, "Set &Ambient Light Color\tA\tSet the ambient light color for the current viewport", NULL, this, ID_SETAMBIENTCOLOR);
  new FXMenuCommand(viewmenu_, "Set &Background Color\tB\tSet the background color for the current viewport", NULL, this, ID_SETBACKCOLOR);
  new FXMenuTitle(menubar, "&View", NULL, viewmenu_);


  // Main components for visualization
  FXSplitter* split = new FXSplitter(this, SPLITTER_REVERSED|SPLITTER_HORIZONTAL|LAYOUT_FILL);

  FXVerticalFrame* gfxframe = new FXVerticalFrame(split, LAYOUT_FILL|FRAME_THICK|FRAME_SUNKEN, 0,0,0,0, 0,0,0,0);
  glvisual_ = new FXGLVisual(app, VISUAL_DOUBLEBUFFER);
  viewer_ = new FXOSGCompositeViewer(gfxframe, glvisual_, this, ID_CANVAS, LAYOUT_FILL);

  FXSplitter* labelsplit = new FXSplitter(split, SPLITTER_REVERSED|SPLITTER_VERTICAL|LAYOUT_FILL);
  FXVerticalFrame* listcontainer = new FXVerticalFrame(labelsplit, LAYOUT_FILL|FRAME_THICK|FRAME_SUNKEN, 0,0,0,0, 0,0,0,0, 0,0);
  new FXLabel(listcontainer, "Scene Elements", NULL, LAYOUT_FILL_X|FRAME_THICK|FRAME_RAISED, 0,0,0,0, 100, 100);
  list_ = new ViewerFoldingList(listcontainer, this, ID_LIST, LAYOUT_FILL|FOLDINGLIST_BROWSESELECT|FOLDINGLIST_ROOT_BOXES|FOLDINGLIST_SHOWS_BOXES|FOLDINGLIST_SHOWS_LINES);
  list_->appendHeader("Name", NULL, 200);
  list_->appendHeader("Type", NULL, 150);

  // Create switcher with property sheets for different node types
  switcher_ = new FXSwitcher(labelsplit, LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);

  // Generic node properties
  FXVerticalFrame* nodeframe = new FXVerticalFrame(switcher_, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  new FXLabel(nodeframe, "Node Properties", NULL, JUSTIFY_LEFT);
  new FXVerticalSeparator(nodeframe, SEPARATOR_LINE|LAYOUT_FILL_X);
  FXMatrix* nodemat = new FXMatrix(nodeframe, 2, MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  new FXLabel(nodemat, "Name:", NULL, JUSTIFY_LEFT|LAYOUT_CENTER_Y);
  nodenames_[NODE_PANEL] = new FXTextField(nodemat, 0, this, ID_NODENAME, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);

  // Switch node properties
  FXVerticalFrame* switchframe = new FXVerticalFrame(switcher_, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  new FXLabel(switchframe, "Switch Node Properties", NULL, JUSTIFY_LEFT);
  new FXVerticalSeparator(switchframe, SEPARATOR_LINE|LAYOUT_FILL_X);
  FXMatrix* switchmat = new FXMatrix(switchframe, 2, MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  new FXLabel(switchmat, "Name:", NULL, JUSTIFY_LEFT|LAYOUT_CENTER_Y);
  nodenames_[SWITCH_PANEL] = new FXTextField(switchmat, 0, this, ID_NODENAME, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  new FXLabel(switchmat, "Active Node:", NULL, JUSTIFY_LEFT);
  FXVerticalFrame* switchlistframe = new FXVerticalFrame(switchmat, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_FILL_COLUMN|LAYOUT_FILL_ROW, 0,0,0,0, 0,0,0,0);
  switchlist_ = new FXList(switchlistframe, this, ID_SWITCHNODE, LIST_MULTIPLESELECT|LAYOUT_FILL_X|LAYOUT_FILL_Y);

  // Transform node properties
  FXVerticalFrame* transframe = new FXVerticalFrame(switcher_, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  new FXLabel(transframe, "PositionAttitudeTransform Node Properties", NULL, JUSTIFY_LEFT);
  new FXVerticalSeparator(transframe, SEPARATOR_LINE|LAYOUT_FILL_X);
  FXMatrix* transmat = new FXMatrix(transframe, 2, MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  new FXLabel(transmat, "Name:", NULL, JUSTIFY_LEFT|LAYOUT_CENTER_Y);
  nodenames_[TRANSFORM_PANEL] = new FXTextField(transmat, 0, this, ID_NODENAME, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);

  new FXLabel(transmat, "Trans:", NULL, JUSTIFY_LEFT|LAYOUT_CENTER_Y);
  FXHorizontalFrame* translateframe = new FXHorizontalFrame(transmat, PACK_UNIFORM_WIDTH|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN, 0,0,0,0, 0,0,0,0);
  transforms_[0] = new FXTextField(translateframe, 0, this, ID_TRANSLATION_X, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  transforms_[1] = new FXTextField(translateframe, 0, this, ID_TRANSLATION_Y, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  transforms_[2] = new FXTextField(translateframe, 0, this, ID_TRANSLATION_Z, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);

  new FXLabel(transmat, "Rot:", NULL, JUSTIFY_LEFT|LAYOUT_CENTER_Y);
  FXHorizontalFrame* rotframe = new FXHorizontalFrame(transmat, PACK_UNIFORM_WIDTH|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN, 0,0,0,0, 0,0,0,0);
  transforms_[3] = new FXTextField(rotframe, 0, this, ID_ROTATION_A, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  transforms_[4] = new FXTextField(rotframe, 0, this, ID_ROTATION_X, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  transforms_[5] = new FXTextField(rotframe, 0, this, ID_ROTATION_Y, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  transforms_[6] = new FXTextField(rotframe, 0, this, ID_ROTATION_Z, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);

  new FXLabel(transmat, "Scale:", NULL, JUSTIFY_LEFT|LAYOUT_CENTER_Y);
  FXHorizontalFrame* scaleframe = new FXHorizontalFrame(transmat, PACK_UNIFORM_WIDTH|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN, 0,0,0,0, 0,0,0,0);
  transforms_[7] = new FXTextField(scaleframe, 0, this, ID_SCALE_X, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  transforms_[8] = new FXTextField(scaleframe, 0, this, ID_SCALE_Y, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  transforms_[9] = new FXTextField(scaleframe, 0, this, ID_SCALE_Z, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);

  // Geometry node properties
  FXVerticalFrame* geomframe = new FXVerticalFrame(switcher_, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  new FXLabel(geomframe, "Geometry Node Properties", NULL, JUSTIFY_LEFT);
  new FXVerticalSeparator(geomframe, SEPARATOR_LINE|LAYOUT_FILL_X);
  FXMatrix* geommat = new FXMatrix(geomframe, 2, MATRIX_BY_COLUMNS|LAYOUT_FILL_X|LAYOUT_FILL_Y, 0,0,0,0, 0,0,0,0);
  new FXLabel(geommat, "Name:", NULL, JUSTIFY_LEFT|LAYOUT_CENTER_Y);
  nodenames_[GEOMETRY_PANEL] = new FXTextField(geommat, 0, this, ID_NODENAME, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_COLUMN);
  new FXLabel(geommat, "Vertices:", NULL, JUSTIFY_LEFT|LAYOUT_CENTER_Y|LAYOUT_FILL_ROW);
  FXVerticalFrame* geomtableframe = new FXVerticalFrame(geommat, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_FILL_COLUMN|LAYOUT_FILL_ROW, 0,0,0,0, 0,0,0,0);
  geomtable_ = new FXTable(geomtableframe, this, ID_GEOMTABLE, TABLE_COL_SIZABLE|TABLE_NO_COLSELECT|TABLE_NO_ROWSELECT|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  geomtable_->setVisibleRows(5);

  // Chore to draw scecne
  app->addChore(this, ID_DRAWCHORE);

  // Add the root to the list
  list_->appendItem(NULL, new FXFoldingItem(makeNodeLabel(root_.get()), 0, 0, root_.get()), true);
}

void ViewerWindow::create()
{
  FXMainWindow::create();

  // Read stored settings from registry
  readRegistry();
}

void ViewerWindow::readRegistry()
{
  loadScenePath_ = getApp()->reg().readStringEntry("SETTINGS", "loadscenepath", loadScenePath_.text());
}

void ViewerWindow::writeRegistry()
{
  getApp()->reg().writeStringEntry("SETTINGS", "loadscenepath", loadScenePath_.text());
}

// Tests if target is an ancestor of node
bool ViewerWindow::isAncestor(osg::Group* target, osg::Node* node)
{
  osg::Node::ParentList& parents = node->getParents();
  osg::Node::ParentList::iterator iter;
  for(iter = parents.begin(); iter != parents.end(); iter++)
  {
    if(*iter == target)
    {
      return true;
    }
    
    if(isAncestor(target, *iter))
    {
      return true;
    }
  }

  return false;
}

void ViewerWindow::removeAllParents(osg::Node* node)
{
  while(node->getNumParents() > 0)
  {
    node->getParent(0)->removeChild(node);
  }
}

void ViewerWindow::setSceneRoot(osg::Node* node)
{
  root_ = node;

  unsigned int i, n = viewer_->getNumViews();
  for(i = 0; i < n; i++)
  {
    viewer_->getView(i)->setSceneData(root_.get());
  }
}

void ViewerWindow::resetCameras()
{
  unsigned int i, n = viewer_->getNumViews();
  for(i = 0; i < n; i++)
  {
    osg::ref_ptr<osgGA::GUIEventAdapter> dummyEvent = viewer_->getView(i)->getEventQueue()->createEvent();
    viewer_->getView(i)->getCameraManipulator()->home(*dummyEvent, *viewer_);
  }
}
void ViewerWindow::beginCreateViewport()
{
  drawingViewport_ = true;
  viewer_->setDefaultCursor(getApp()->getDefaultCursor(DEF_CROSSHAIR_CURSOR));
  viewer_->setDragCursor(getApp()->getDefaultCursor(DEF_CROSSHAIR_CURSOR));
}

void ViewerWindow::endCreateViewport()
{
  viewer_->setDragCursor(getApp()->getDefaultCursor(DEF_ARROW_CURSOR));
  viewer_->setDefaultCursor(getApp()->getDefaultCursor(DEF_ARROW_CURSOR));
  drawingViewport_ = false;
}

FXint ViewerWindow::fixCreateViewportX(FXint x)
{
  if(x < 0)
  {
    x = 0;
  }
  else if(x > viewer_->getWidth())
  {
    x = viewer_->getWidth();
  }

  return x;
}

FXint ViewerWindow::fixCreateViewportY(FXint y)
{
  if(y < 0)
  {
    y = 0;
  }
  else if(y > viewer_->getHeight())
  {
    y = viewer_->getHeight();
  }

  return viewer_->getHeight() - y;
}

void ViewerWindow::addView(const std::string& name, bool fixed, double x, double y, double w, double h)
{
  // Main view
  osg::ref_ptr<osgViewer::View> view = new osgViewer::View();
  osg::ref_ptr<osg::Camera> camera = new osg::Camera();

  camera->setViewport(fixed ? new osg::Viewport(x, y, w, h) : new RatioViewport(x, y, w, h, viewer_->getWidth(), viewer_->getHeight()));
  camera->setClearColor(osg::Vec4(0.2f, 0.2f, 0.4f, 1.0f));
  camera->setGraphicsContext(viewer_);
  camera->setProjectionMatrixAsPerspective(45.0, camera->getViewport()->width()/camera->getViewport()->height(), 0.1, 1000.0);

  view->setCamera(camera.get());

  osg::ref_ptr<osgGA::TrackballManipulator> manipulator = new osgGA::TrackballManipulator();
  view->setCameraManipulator(manipulator.get());
  view->addEventHandler(new osgGA::StateSetManipulator(view->getCamera()->getOrCreateStateSet()));
  view->setSceneData(root_.get());

  view->setName(name);

  viewer_->addView(view.get());

  viewer_->setUpBorderRenderingSupport();
}

void ViewerWindow::removeView(unsigned int index)
{
  if(index < viewer_->getNumViews())
  {
    osg::ref_ptr<osgViewer::View> view = viewer_->getView(index);
    viewer_->removeView(view.get());
    view->setSceneData(NULL);
    view->getCamera()->setViewport(NULL);
    view->setCamera(NULL);
  }
}

bool ViewerWindow::loadModel(const char* name)
{
  // load the scene.
  osg::ref_ptr<osg::Node> loadedModel = osgDB::readNodeFile(name);

  if (!loadedModel)
  {
    return false;
  }

  osgUtil::Optimizer optimizer;
  optimizer.optimize(loadedModel.get());

  setSceneRoot(loadedModel.get());

  populateList();

  currentFile_ = name;
  dirty_ = false;

  return true;
}

bool ViewerWindow::saveModel(const char* name)
{
  if(osgDB::writeNodeFile(*root_, name))
  {
    dirty_ = false;

    return true;
  }

  return false;
}

bool ViewerWindow::open()
{
  FXFileDialog dialog(this, "Load Scene Data");

  dialog.setSelectMode(SELECTFILE_EXISTING);
  dialog.setPatternList("OSG Files (*.osg|*.osga)\nAll Files (*)");
  dialog.setDirectory(loadScenePath_);

  for(;;)
  {
    if(dialog.execute(PLACEMENT_OWNER))
    {
      FXbool success;
      FXString file = dialog.getFilename();
      FXString path = dialog.getDirectory();

      // Check file
      if(!FXStat::isReadable(file))
      {
        FXMessageBox::error(this, MBOX_OK, "Load Error", "The file [%s] is read protected.  ", file.text());
        continue;
      }

      // Load file
      getApp()->beginWaitCursor();
      success = loadModel(file.text());
      getApp()->endWaitCursor();

      if(!success)
      {
        FXMessageBox::error(this, MBOX_OK, "Load Error", "Could not load file [%s].  \nThe system reported the following error:\n\n%s\n", file.text(), strerror(errno));
      }
      else
      {
        loadScenePath_ = path;
      }
    }
    else
    {
      return false;
    }

    // If not continued, loop is canceled
    break;
  }

  return true;
}

bool ViewerWindow::saveAs()
{
  FXFileDialog dialog(this, "Load Scene Data");

  dialog.setSelectMode(SELECTFILE_ANY);
  dialog.setPatternList("OSG File (*.osg)\nOSG Archive File (*.osga)");

  if(currentFile_.empty())
  {
    dialog.setDirectory(loadScenePath_);
  }
  else
  {
    dialog.setFilename(currentFile_);
  }

  for(;;)
  {
    if(dialog.execute(PLACEMENT_OWNER))
    {
      FXbool success;
      FXString path = dialog.getDirectory();
      FXString file = dialog.getFilename();

      // Force file extenstion
      if(FXPath::extension(file).empty())
      {
        file += dialog.getPattern().after('*');
      }

      if(FXStat::exists(file) && MBOX_CLICKED_NO == FXMessageBox::error(this, MBOX_YES_NO, "File Already Exists", "Overwrite existing file [%s]?  ", file.text()))
      {
        continue;
      }

      // Save file
      getApp()->beginWaitCursor();
      success = saveModel(file.text());
      getApp()->endWaitCursor();

      if(!success)
      {
        FXMessageBox::error(this, MBOX_OK, "Save Error", "Could not save file [%s].  \nThe system reported the following error:\n\n%s\n", file.text(), strerror(errno));
      }
      else
      {
        loadScenePath_ = path;
        currentFile_ = file;
      }
    }
    else
    {
      // Clicked cancel at file dialog
      return false;
    }

    // If not continued, loop is canceled
    break;
  }

  return true;
}

FXString ViewerWindow::makeNodeLabel(osg::Node* node)
{
  FXString type = typeid(*node).name();
  FXString name = node->getName().c_str();
  return FXStringFormat("%s\t%s", name.empty() ? "<anonymous>" : name.text(), type.after(' ').text());
}

FXString ViewerWindow::makeDrawableLabel(osg::Drawable* drawable)
{
  FXString type = typeid(*drawable).name();
  FXString name = drawable->getName().c_str();
  return FXStringFormat("%s\t%s", name.empty() ? "<anonymous>" : name.text(), type.after(' ').text());
}

FXString ViewerWindow::makePrimitiveSetLabel(osg::PrimitiveSet* set)
{
  FXString type = typeid(*set).name();
  FXString name = "UNKNOWN";

  switch(set->getMode())
  {
  case osg::PrimitiveSet::POINTS:
    name = "POINTS";
    break;
  case osg::PrimitiveSet::LINES:
    name = "LINES";
    break;
  case osg::PrimitiveSet::LINE_STRIP:
    name = "LINE_STRIP";
    break;
  case osg::PrimitiveSet::LINE_LOOP:
    name = "LINE_LOOP";
    break;
  case osg::PrimitiveSet::TRIANGLES:
    name = "TRIANGLES";
    break;
  case osg::PrimitiveSet::TRIANGLE_STRIP:
    name = "TRIANGLE_STRIP";
    break;
  case osg::PrimitiveSet::TRIANGLE_FAN:
    name = "TRIANGLE_FAN";
    break;
  case osg::PrimitiveSet::QUADS:
    name = "QUADS";
    break;
  case osg::PrimitiveSet::QUAD_STRIP:
    name = "QUAD_STRIP";
    break;
  case osg::PrimitiveSet::POLYGON:
    name = "POLYGON";
    break;
  }

  return FXStringFormat("%s\t%s", name.text(), type.after(' ').text());
}

void ViewerWindow::populateList()
{
  list_->clearItems();
  process(root_.get(), NULL);
}

void ViewerWindow::process(osg::Node* node, FXFoldingItem* parent)
{
  osg::Geode *geode = dynamic_cast<osg::Geode*>(node);
  if(geode)
  {
    FXFoldingItem* item = new FXFoldingItem(makeNodeLabel(geode), 0, 0, geode);
    item->setExpanded(true);
    list_->appendItem(parent, item, true);
    
    processGeode(geode, item);
  }
  else
  {
    osg::Group *gp = dynamic_cast<osg::Group*>(node);

    if(gp)
    {
      FXFoldingItem* item = new FXFoldingItem(makeNodeLabel(gp), 0, 0, gp);
      item->setExpanded(true);
      list_->appendItem(parent, item, true);

      for (unsigned int i = 0; i < gp->getNumChildren(); i++)
      {
        process(gp->getChild(i), item);
      }
    }
    else
    {
      // Unknown node type - can not process further
      FXFoldingItem* item = new FXFoldingItem(makeNodeLabel(node), 0, 0, node);
      list_->appendItem(parent, item, true);
    }
  }
}

void ViewerWindow::processGeode(osg::Geode* geode, FXFoldingItem* parent)
{
  for (unsigned int i=0; i<geode->getNumDrawables(); i++)
  {
    osg::Drawable *drawable = geode->getDrawable(i);
    FXFoldingItem* item = new FXFoldingItem(makeDrawableLabel(drawable), 0, 0, drawable);
    list_->appendItem(parent, item, true);

    osg::Geometry *geom=dynamic_cast<osg::Geometry*>(drawable);
    if(geom)
    {
      for (unsigned int j = 0; j < geom->getNumPrimitiveSets(); j++)
      {
        osg::PrimitiveSet* set = geom->getPrimitiveSet(j);
        FXFoldingItem* subitem = new FXFoldingItem(makePrimitiveSetLabel(set), 0, 0, set);
        subitem->setEnabled(false);
        list_->appendItem(item, subitem, true);
      }
    }
  }
}

long ViewerWindow::onUpdTitleBar(FXObject*, FXSelector, void*)
{
  FXString title = TITLE;
  if(!currentFile_.empty())
  {
    FXString path = FXPath::directory(currentFile_);
    if(!path.empty())
    {
      title += FXStringFormat(" - [%s - %s]", FXPath::name(currentFile_).text(), path.text());
    }
    else
    {
      title += FXStringFormat(" - [%s]", FXPath::name(currentFile_).text());
    }

    if(dirty_)
    {
      title += "*";
    }
  }

  handle(this, FXSEL(SEL_COMMAND, ID_SETSTRINGVALUE), (void*)&title);

  return 1;
}

long ViewerWindow::onCmdQuit(FXObject*, FXSelector, void*)
{
  // First check for modifications to current scene
  if(dirty_)
  {
    FXuint result = FXMessageBox::error(this, MBOX_YES_NO_CANCEL, "Save Changes", "Save changes to current scene before proceeding?  ");

    if(result == MBOX_CLICKED_YES)
    {
      // Save file if modified; if save is canceled the load is aborted
      if(!saveAs())
        return 1;
    }
    else if( result == MBOX_CLICKED_CANCEL)
    {
      // Cancel the quit command
      return 1;
    }
  }

  writeRegistry();
  getApp()->exit(0);
  return 1;
}

long ViewerWindow::onCmdLoadScene(FXObject*, FXSelector, void*)
{
  // First check for modifications to current scene
  if(dirty_)
  {
    FXuint result = FXMessageBox::error(this, MBOX_YES_NO_CANCEL, "Save Changes", "Save changes to current scene before proceeding?  ");

    if(result == MBOX_CLICKED_YES)
    {
      // Save file if modified; if save is canceled the load is aborted
      if(!saveAs())
        return 1;
    }
    else if( result == MBOX_CLICKED_CANCEL)
    {
      // Cancel the quit command
      return 1;
    }
  }

  open();

  return 1;
}

long ViewerWindow::onCmdSaveScene(FXObject*, FXSelector, void*)
{
  FXbool success;

  // Save file
  getApp()->beginWaitCursor();
  success = saveModel(currentFile_.text());
  getApp()->endWaitCursor();

  if(!success)
  {
    FXMessageBox::error(this, MBOX_OK, "Save Error", "Could not save file [%s].  \nThe system reported the following error:\n\n%s\n", currentFile_.text(), strerror(errno));
  }

  return 1;
}

long ViewerWindow::onUpdSaveScene(FXObject* sender, FXSelector, void*)
{
  FXuint msg = dirty_ ? ID_ENABLE : ID_DISABLE;
  sender->handle(this, FXSEL(SEL_COMMAND, msg), NULL);
  return 1;
}

long ViewerWindow::onCmdSaveSceneAs(FXObject*, FXSelector, void*)
{
  saveAs();
  return 1;
}

long ViewerWindow::onCmdManageViewports(FXObject*, FXSelector, void*)
{
  ViewportDialog dialog(this, this, viewer_);
  dialog.execute(PLACEMENT_OWNER);
  return 1;
}

long ViewerWindow::onCmdDrawViewport(FXObject*, FXSelector, void*)
{
  beginCreateViewport();
  return 1;
}

long ViewerWindow::onCmdResetCameras(FXObject*, FXSelector, void*)
{
  resetCameras();
  return 1;
}

long ViewerWindow::onCmdSetBackColor(FXObject*, FXSelector, void*)
{
  // Get the old color
  osg::Vec4 color = viewer_->getCameraWithFocus()->getClearColor();

  // Select a new color
  FXColorDialog dlg(this, "Background Color");

  // Show modal dialog with current color as default
  dlg.setRGBA(color.asRGBA());
  if(dlg.execute())
  {
    FXColor newcolor = dlg.getRGBA();
    viewer_->getCameraWithFocus()->setClearColor(osg::Vec4(FXREDVAL(newcolor)/255.0, FXGREENVAL(newcolor)/255.0, FXBLUEVAL(newcolor)/255.0, FXALPHAVAL(newcolor)/255.0));
  }

  return 1;
}

long ViewerWindow::onCmdSetAmbientColor(FXObject*, FXSelector, void*)
{
  osgUtil::SceneView* sv = viewer_->getSceneView(viewer_->getViewWithFocus());
  osg::Vec4 color = sv->getLight()->getAmbient();

  // Select a new color
  FXColorDialog dlg(this, "Background Color");

  // Show modal dialog with current color as default
  dlg.setRGBA(color.asRGBA());
  if(dlg.execute())
  {
    FXColor newcolor = dlg.getRGBA();
    sv->getLight()->setAmbient(osg::Vec4(FXREDVAL(newcolor)/255.0, FXGREENVAL(newcolor)/255.0, FXBLUEVAL(newcolor)/255.0, FXALPHAVAL(newcolor)/255.0));
  }

  return 1;
}

long ViewerWindow::onLeftBtnPress(FXObject*, FXSelector, void* data)
{
  if(drawingViewport_)
  {
    FXEvent *event = (FXEvent*)data;

    drawStart_[0] = event->win_x;
    drawStart_[1] = event->win_y;

    return 1;
  }

  return 0;
}

long ViewerWindow::onLeftBtnRelease(FXObject*, FXSelector, void* data)
{
  if(drawingViewport_)
  {
    FXEvent *event = (FXEvent*)data;
    FXint x1 = fixCreateViewportX(drawStart_[0]);
    FXint y1 = fixCreateViewportY(drawStart_[1]);
    FXint x2 = fixCreateViewportX(event->win_x);
    FXint y2 = fixCreateViewportY(event->win_y);

    // Only create a viewport if width and height are greater than 0
    if((x1 != x2) && (y1 != y2))
    {
      double x, y, w, h;
      double width = viewer_->getWidth();
      double height = viewer_->getHeight();

      if(x1 < x2)
      {
        x = x1;
        w = x2 - x1;
      }
      else
      {
        x = x2;
        w = x1 - x2;
      }

      if(y1 < y2)
      {
        y = y1;
        h = y2 - y1;
      }
      else
      {
        y = y2;
        h = y1 - y2;
      }

      addView("NewViewport", false, x/width, y/height, w/width, h/height);
    }

    endCreateViewport();

    return 1;
  }

  return 0;
}

long ViewerWindow::onMotion(FXObject*, FXSelector, void*)
{
  if(drawingViewport_)
  {
    return 1;
  }

  return 0;
}

long ViewerWindow::onCmdList(FXObject*, FXSelector, void* data)
{
  currentItem_ = (FXFoldingItem*)(data);
  osg::ref_ptr<osg::Node> currentNode = (osg::Node*)currentItem_->getData();

  FXString name = currentNode->getName().c_str();
  if(name.empty())
  {
    name = "<anonymous>";
  }

  // Determine type
  if(typeid(*currentNode) == typeid(osg::Switch))
  {
    switcher_->setCurrent(SWITCH_PANEL);
    nodenames_[SWITCH_PANEL]->setText(name);

    osg::ref_ptr<osg::Switch> node = dynamic_cast<osg::Switch*>(currentNode.get());

    // Populate the list
    switchlist_->clearItems();

    unsigned int i, n = node->getNumChildren();
    for(i = 0; i < n; i++)
    {
      FXString s = node->getChild(i)->getName().c_str();
      switchlist_->appendItem(s.empty() ? "<anonymous>" : s);

      if(node->getValue(i))
      {
        switchlist_->selectItem(i);
      }
    }
  }
  else if(typeid(*currentNode) == typeid(osg::PositionAttitudeTransform))
  {
    switcher_->setCurrent(TRANSFORM_PANEL);
    nodenames_[TRANSFORM_PANEL]->setText(name);

    osg::ref_ptr<osg::PositionAttitudeTransform> node = dynamic_cast<osg::PositionAttitudeTransform*>(currentNode.get());

    osg::Vec3d trans = node->getPosition();
    osg::Quat rot = node->getAttitude();
    osg::Vec3d scale = node->getScale();

    transforms_[0]->setText(FXStringVal(trans.x()));
    transforms_[1]->setText(FXStringVal(trans.y()));
    transforms_[2]->setText(FXStringVal(trans.z()));

    double a, x, y, z;
    rot.getRotate(a, x, y, z);
    transforms_[3]->setText(FXStringVal(a));
    transforms_[4]->setText(FXStringVal(x));
    transforms_[5]->setText(FXStringVal(y));
    transforms_[6]->setText(FXStringVal(z));

    transforms_[7]->setText(FXStringVal(scale.x()));
    transforms_[8]->setText(FXStringVal(scale.y()));
    transforms_[9]->setText(FXStringVal(scale.z()));
  }
  else if(typeid(*currentNode) == typeid(osg::Geometry))
  {
    switcher_->setCurrent(GEOMETRY_PANEL);
    nodenames_[GEOMETRY_PANEL]->setText(name);

    osg::ref_ptr<osg::Geometry> node = dynamic_cast<osg::Geometry*>(currentNode.get());
    osg::Vec3Array* verts = dynamic_cast<osg::Vec3Array*>(node->getVertexArray());

    unsigned int i, n = verts->getNumElements();
    geomtable_->setTableSize(n, 3);

    geomtable_->setColumnText(0, "X");
    geomtable_->setColumnText(1, "Y");
    geomtable_->setColumnText(2, "Z");

    for(i = 0; i < n; i++)
    {
      geomtable_->setRowText(i, FXStringVal(i + 1));
      geomtable_->setItemText(i, 0, FXStringVal((*verts)[i].x()));
      geomtable_->setItemText(i, 1, FXStringVal((*verts)[i].y()));
      geomtable_->setItemText(i, 2, FXStringVal((*verts)[i].z()));
    }

    FXuint tablewidth = geomtable_->getWidth();
    FXuint rowheaderwidth = geomtable_->getRowHeader()->getDefaultWidth();
    FXuint colwidth = (tablewidth - rowheaderwidth) / 3;

    rowheaderwidth +=(tablewidth - rowheaderwidth) % 3;

    geomtable_->setRowHeaderWidth(rowheaderwidth);
    geomtable_->setColumnWidth(0, colwidth);
    geomtable_->setColumnWidth(1, colwidth);
    geomtable_->setColumnWidth(2, colwidth);
  }
  else
  {
    switcher_->setCurrent(NODE_PANEL);
    nodenames_[NODE_PANEL]->setText(name);
  }

  return 1;
}

long ViewerWindow::onCmdNodeName(FXObject* sender, FXSelector, void*)
{
  FXString name;
  sender->handle(this, FXSEL(SEL_COMMAND, ID_GETSTRINGVALUE), (void*)&name);

  list_->setItemText(currentItem_, name);

  osg::ref_ptr<osg::Node> currentNode = (osg::Node*)currentItem_->getData();
  currentNode->setName(name.text());

  dirty_ = true;

  return 1;
}

long ViewerWindow::onCmdSwitchNode(FXObject*, FXSelector, void*)
{
  return 1;
}

long ViewerWindow::onCmdPositionAttitudeTransform(FXObject*, FXSelector, void*)
{
  return 1;
}

long ViewerWindow::onCmdGeometryTable(FXObject*, FXSelector, void* data)
{
  FXTableRange* range = (FXTableRange*)data;

  // The vertex range is sent before the cells are updated
  // The range is stored and processed by a chore
  memcpy(&vertexRange_, range, sizeof(vertexRange_));

  getApp()->addChore(this, ID_GEOMCHORE);

  return 1;
}

long ViewerWindow::onGeometryChore(FXObject*, FXSelector, void*)
{
  // Update the vertices for the current geometry
  osg::ref_ptr<osg::Geometry> node = dynamic_cast<osg::Geometry*>((osg::Node*)list_->getItemData(currentItem_));

  if(node != NULL)
  {
    // Make a copy, the change will stay local to this geometry object if the vertices are shared
    osg::Vec3Array* oldverts = dynamic_cast<osg::Vec3Array*>(node->getVertexArray());
    if(oldverts != NULL)
    {
      osg::Vec3Array* verts = new osg::Vec3Array(*oldverts);
      int i, n = vertexRange_.to.row;
      for(i = vertexRange_.fm.row; i <= n + 1; i++)
      {
        osg::Vec3 vert = (*verts)[i];

        int j, m = vertexRange_.to.col;
        for(j = vertexRange_.fm.col; j <= m; j++)
        {
          vert[j] = FXDoubleVal(geomtable_->getItemText(i, j));
          dirty_ = true;
        }
        
        (*verts)[i].set(vert);
      }

      if(dirty_)
      {
        node->setVertexArray(verts);
      }
    }
  }

  return 1;
}

long ViewerWindow::onDrawChore(FXObject*, FXSelector, void*)
{
  // Reschedule timer (manual reset)
  getApp()->addChore(this, ID_DRAWCHORE);

  // Request update from window manager
  viewer_->draw();

  return 1;
}

long ViewerWindow::onDNDDrop(FXObject*, FXSelector, void* ptr)
{
  FXEvent* event = (FXEvent*)ptr;
  FXuchar* data;
  FXuint len;

  // Attempt to reparent dragged item to item under mouse at time of drop
  if(getDNDData(FROM_DRAGNDROP, ViewerFoldingList::foldingItemType, data, len))
  {
    FXFoldingItem* child;

    memcpy((FXchar*)&child, (FXchar*)data, len);

    // Only process the child if it came from this list (child is equal to the currentItem_)
    if(child == currentItem_)
    {
      FXFoldingItem* newparent = list_->getItemAt(event->win_x, event->win_y);

      if(newparent != NULL)
      {
        // Check for compatibility with new parent - Geometry can not be a parent and can only be parented to Geode
        osg::ref_ptr<osg::Geometry> newgeometry = dynamic_cast<osg::Geometry*>((osg::Node*)list_->getItemData(newparent));
        if(newgeometry != NULL)
        {
          FXMessageBox::error(this, MBOX_OK, "Drag and Drop Error", "The Geometry Node object type can not be a parent to another Node.  ");
        }
        else
        {
          FXFoldingItem* oldparent = currentItem_->getParent();
          osg::ref_ptr<osg::Node> node = (osg::Node*)list_->getItemData(currentItem_);
          osg::ref_ptr<osg::Node> newnode = (osg::Node*)list_->getItemData(newparent);

          // Is the new parent a geode?
          osg::ref_ptr<osg::Geode> newgeode = dynamic_cast<osg::Geode*>(newnode.get());
          if(newgeode != NULL)
          {
            // Is the child a geometry node?
            osg::ref_ptr<osg::Geometry> geometry = dynamic_cast<osg::Geometry*>(node.get());
            if(geometry == NULL)
            {
              FXMessageBox::error(this, MBOX_OK, "Drag and Drop Error", "The Geode object type can only be a parent to a Geometry Node.  ");
            }
            else
            {
              // Remove node from original parent (if it is not root item)
              if(oldparent != NULL)
              {
                osg::ref_ptr<osg::Geode> oldgeode = dynamic_cast<osg::Geode*>((osg::Node*)list_->getItemData(oldparent));
                oldgeode->removeDrawable(geometry.get());
              }
              else
              {
                // The root has been moved to a parent which is the new root
                removeAllParents(newnode.get());
                setSceneRoot(newnode.get());
              }

              newgeode->addDrawable(geometry.get());
              list_->moveItem(NULL, newparent, currentItem_);

              dirty_ = true;
            }
          }
          else
          {
            // Is it a group node
            osg::ref_ptr<osg::Group> newgroup = dynamic_cast<osg::Group*>(newnode.get());
            if(newgroup != NULL)
            {
              // The child of a group should not be a geometry node
              osg::ref_ptr<osg::Geometry> geometry = dynamic_cast<osg::Geometry*>(node.get());
              if(geometry != NULL)
              {
                FXMessageBox::error(this, MBOX_OK, "Drag and Drop Error", "The Geometry Node object type can only be a child of a Geode object.  ");
              }
              else
              {
                // Remove node from original parent (if it is not root item)
                if(oldparent != NULL)
                {
                  osg::ref_ptr<osg::Group> oldgroup = dynamic_cast<osg::Group*>((osg::Node*)list_->getItemData(oldparent));

                  // If the node is an ancestor of the newnode, the parents must be exchanged

                  osg::ref_ptr<osg::Group> target = dynamic_cast<osg::Group*>(node.get());
                  if(target != NULL && isAncestor(target.get(), newgroup.get()))
                  {
                    removeAllParents(newnode.get());
                    oldgroup->addChild(newnode.get());
                  }

                  oldgroup->removeChild(node.get());
                }
                else
                {
                  // The root has been moved to a parent which is the new root
                  removeAllParents(newnode.get());
                  setSceneRoot(newnode.get());

                  // Make the new parent the root item
//                  list_->moveItem(NULL, NULL, newparent);
                }

                newgroup->addChild(node.get());

//                list_->moveItem(NULL, newparent, currentItem_);
                populateList();

                dirty_ = true;
              }
            }
            else
            {
              // It is an unsupported parent type
              FXMessageBox::error(this, MBOX_OK, "Drag and Drop Error", "The Node type for the newly selected parent is not supported by the reparenting operation.  ");
            }
          }
        }
      }
    }

    FXFREE(&data);
  }

  return 1;
}

long ViewerWindow::onDNDRequest(FXObject*, FXSelector, void* ptr)
{
  FXEvent* event = (FXEvent*)ptr;

  if(event->target == ViewerFoldingList::foldingItemType)
  {
    if(currentItem_ != NULL)
    {
      FXuchar *data; 
      FXuint len = sizeof(FXFoldingItem*);

      FXMALLOC(&data, FXuchar, len);
      memcpy((FXchar*)data, (FXchar*)&currentItem_, len);

      setDNDData(FROM_DRAGNDROP, ViewerFoldingList::foldingItemType, data, len);

      return 1;
    }
  }

  return 0;
}

ViewerWindow::~ViewerWindow()
{
  while(viewer_->getNumViews() > 0)
  {
    removeView(0);
  }

  delete filemenu_;
  delete viewmenu_;
  delete glvisual_;
}
