#ifndef RATIOVIEWPORT_H
#define RATIOVIEWPORT_H

#include <osg/Viewport>

namespace osgFOX
{

class RatioViewport : public osg::Viewport
{
protected:
  value_type rx_;               // The x ration
  value_type ry_;               // The y ration
  value_type rwidth_;           // The width ration
  value_type rheight_;          // The height ratio
public:
  RatioViewport() : rx_(0.0), ry_(0.0), rwidth_(1.0), rheight_(1.0) { }

  RatioViewport(value_type rx,value_type ry,value_type rwidth,value_type rheight) : rx_(rx), ry_(ry), rwidth_(rwidth), rheight_(rheight)
  {
    _x = _width * rx_;
    _y = _height * ry_;
    _width *= rwidth_;
    _height *= rheight_;
  }

  RatioViewport(value_type rx,value_type ry,value_type rwidth,value_type rheight,value_type width,value_type height) : osg::Viewport(width * rx, height * ry, width * rwidth, height * rheight), rx_(rx), ry_(ry), rwidth_(rwidth), rheight_(rheight) { }

  /** Copy constructor using CopyOp to manage deep vs shallow copy. */
  RatioViewport(const RatioViewport& vp,const osg::CopyOp& copyop=osg::CopyOp::SHALLOW_COPY) : osg::Viewport(vp,copyop), rx_(vp.rx_), ry_(vp.ry_), rwidth_(vp.rwidth_), rheight_(vp.rheight_) {}

  value_type& rx() { return rx_; }

  value_type rx() const { return rx_; }

  value_type& ry() { return ry_; }

  value_type ry() const { return ry_; }

  value_type& rwidth() { return rwidth_; }

  value_type rwidth() const { return rwidth_; }

  value_type& rheight() { return rheight_; }

  value_type rheight() const { return rheight_; }

  inline void resize(value_type width,value_type height)
  {
    _x = width * rx_;
    _y = height * ry_;
    _width = width * rwidth_;
    _height = height * rheight_;
  }

  inline void setViewportRatio(value_type rx,value_type ry,value_type rwidth,value_type rheight)
  {
    rx_ = rx;
    ry_ = ry;
    rwidth_ = rwidth;
    rheight_ = rheight;
  }
};

}

#endif
