#include "RatioViewport.h"
#include "ViewBorderOperation.h"
#include "FXOSGCompositeViewer.h"

using namespace osgFOX;

FXDEFMAP(FXOSGCompositeViewer) FXOSGCompositeViewerMap[] = {
  FXMAPFUNC(SEL_CONFIGURE,0,FXOSGCompositeViewer::onConfigure),
  FXMAPFUNC(SEL_PAINT, 0, FXOSGCompositeViewer::onPaint)
};

FXIMPLEMENT(FXOSGCompositeViewer, GraphicsWindowFOX, FXOSGCompositeViewerMap, ARRAYNUMBER(FXOSGCompositeViewerMap))

FXOSGCompositeViewer::FXOSGCompositeViewer(FXComposite *p,FXGLVisual *vis,FXObject *tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h)
: GraphicsWindowFOX(p, vis, tgt, sel, opts, x, y, w, h),
  borderWidth_(0),
  borderColor_(0xffffffff),
  focusWidth_(0),
  focusColor_(0xff00ffff)
{
  osg::GraphicsContext::setState(new osg::State);
  osg::GraphicsContext::getState()->setContextID(osg::GraphicsContext::createNewContextID());
  osgViewer::CompositeViewer::setEventQueue(GraphicsWindowFOX::getEventQueue());
}

FXOSGCompositeViewer::FXOSGCompositeViewer(FXComposite *p,FXGLVisual *vis,FXGLCanvas *sharegroup,FXObject *tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h)
: GraphicsWindowFOX(p, vis, tgt, sel, opts, x, y, w, h),
  borderWidth_(0),
  borderColor_(0xffffffff),
  focusWidth_(0),
  focusColor_(0xff00ffff)
{
  osg::GraphicsContext::setState(new osg::State);
  osg::GraphicsContext::getState()->setContextID(osg::GraphicsContext::createNewContextID());
  osgViewer::CompositeViewer::setEventQueue(GraphicsWindowFOX::getEventQueue());
}

FXOSGCompositeViewer::~FXOSGCompositeViewer()
{
  osg::GraphicsContext::decrementContextIDUsageCount(osg::GraphicsContext::getState()->getContextID());
}

void FXOSGCompositeViewer::create()
{
  GraphicsWindowFOX::create();
}

long FXOSGCompositeViewer::onConfigure(FXObject *,FXSelector,void *data)
{
  if(target && target->handle(this, MKUINT(message, SEL_CONFIGURE), data)) return 1;

  GraphicsWindowFOX::getEventQueue()->windowResize(0, 0, getWidth(), getHeight());

  Views::iterator iter;
  for(iter = _views.begin(); iter != _views.end(); ++iter)
  {
    if((*iter)->getCamera()->getViewport())
    {
      osg::Viewport* vp = (*iter)->getCamera()->getViewport();
      double previousAspectRatio = static_cast<double>(vp->width()) / static_cast<double>(vp->height());

      // Resize the viewport
      RatioViewport* rvp = dynamic_cast<RatioViewport*>(vp);
      if(rvp != NULL)
      {
        rvp->resize(getWidth(), getHeight());
        (*iter)->getEventQueue()->windowResize(rvp->x(), rvp->y(), rvp->width(), rvp->height());
      }

      double newAspectRatio = static_cast<double>(vp->width()) / static_cast<double>(vp->height());

      // If aspect ratio adjusted change the project matrix to suit
      if (previousAspectRatio != newAspectRatio)
      {
        osg::Matrixd& pm = (*iter)->getCamera()->getProjectionMatrix();
        bool orthographicCamera = (pm(0,3)==0.0) && (pm(1,3)==0.0) && (pm(2,3)==0.0) && (pm(3,3)==1.0);
        if (orthographicCamera)
        {
          double left, right, bottom, top, zNear, zFar;
          (*iter)->getCamera()->getProjectionMatrixAsOrtho(left, right, bottom, top, zNear, zFar);

          double mid = (right+left)*0.5;
          double halfWidth = (right-left)*0.5;
          left = mid - halfWidth * (newAspectRatio/previousAspectRatio);
          right = mid + halfWidth * (newAspectRatio/previousAspectRatio);
          (*iter)->getCamera()->setProjectionMatrixAsOrtho(left, right, bottom, top, zNear, zFar);
        }
        else
        {
          double left, right, bottom, top, zNear, zFar;
          (*iter)->getCamera()->getProjectionMatrixAsFrustum(left, right, bottom, top, zNear, zFar);

          double mid = (right+left)*0.5;
          double halfWidth = (right-left)*0.5;
          left = mid - halfWidth * (newAspectRatio/previousAspectRatio);
          right = mid + halfWidth * (newAspectRatio/previousAspectRatio);
          (*iter)->getCamera()->setProjectionMatrixAsFrustum(left, right, bottom, top, zNear, zFar);
        }
      }
    }
  }

  return 1;
}

long FXOSGCompositeViewer::onPaint(FXObject *,FXSelector,void *data)
{
  draw();
  return 1;
}

void FXOSGCompositeViewer::draw()
{
  // Render OSG frame
  frame();
}

void FXOSGCompositeViewer::drawViewportBorder(osgViewer::View* view)
{
  if(view == NULL)
    return;

  glViewport(0, 0, getWidth(), getHeight());
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D(0.0, getWidth(), 0.0, getHeight());
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  glPushAttrib(GL_ENABLE_BIT);

  glDisable(GL_DEPTH_TEST);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_STENCIL_TEST);

  if(view == getViewWithFocus() && focusWidth_ > 0)
  {
    glLineWidth(focusWidth_);
    glColor4d(FXREDVAL(focusColor_)/255.0, FXGREENVAL(focusColor_)/255.0, FXBLUEVAL(focusColor_)/255.0, FXALPHAVAL(focusColor_)/255.0);

    osg::Viewport* vp = view->getCamera()->getViewport();
    glBegin(GL_LINE_LOOP);
      glVertex2d(vp->x(), vp->y());
      glVertex2d(vp->x() + vp->width(), vp->y());
      glVertex2d(vp->x() + vp->width(), vp->y() + vp->height());
      glVertex2d(vp->x(), vp->y() + vp->height());
    glEnd();
  }
  else if(borderWidth_ > 0)
  {
    glLineWidth(borderWidth_);
    glColor4d(FXREDVAL(borderColor_)/255.0, FXGREENVAL(borderColor_)/255.0, FXBLUEVAL(borderColor_)/255.0, FXALPHAVAL(borderColor_)/255.0);

    osg::Viewport* vp = view->getCamera()->getViewport();
    glBegin(GL_LINE_LOOP);
      glVertex2d(vp->x(), vp->y());
      glVertex2d(vp->x() + vp->width(), vp->y());
      glVertex2d(vp->x() + vp->width(), vp->y() + vp->height());
      glVertex2d(vp->x(), vp->y() + vp->height());
    glEnd();
  }

  glPopAttrib();

  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
}

void FXOSGCompositeViewer::setUpBorderRenderingSupport()
{
  Contexts contexts;
  getContexts(contexts);

  for(Contexts::iterator gcitr = contexts.begin();
    gcitr != contexts.end();
    ++gcitr)
  {
    unsigned int pos = 1;
    osg::GraphicsContext* gc = *gcitr;
    osg::GraphicsContext::Cameras& cameras = gc->getCameras();

    osg::GraphicsContext::OperationQueue& currentops_ = gc->getOperationsQueue();

   for(osg::GraphicsContext::Cameras::iterator citr = cameras.begin();
        citr != cameras.end();
        ++citr)
    {
      osg::Camera* camera = *citr;
      osgViewer::View* view = dynamic_cast<osgViewer::View*>(camera->getView());

      osg::GraphicsContext::OperationQueue::iterator op = currentops_.begin();
      for(unsigned int i = 0; i < pos; i++)
        ++op;
      currentops_.insert(op, new ViewBorderOperation(view, this));
      pos += 2;
    }
  }
}

void FXOSGCompositeViewer::removeView(osgViewer::View* view)
{
  Views::iterator iter;
  for(iter = _views.begin(); iter != _views.end(); ++iter)
  {
    if((*iter) == view)
    {
      _cameraSceneViewMap.erase(view->getCamera());
      _views.erase(iter);
      break;
    }
  }
}
